# SPDX-FileCopyrightText: Coypright © 2024 Shooting Soul Ventures, LLC <jg@shootingsoul.com>
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from typing import List


@dataclass
class Sweep:
    """
    Replace the value in a config with a Sweep object with a list of values to sweep over

    values - list of values to sweep over
    Priority - priorty relative to other sweep objects.
               Higher priorty items group their values together in the combinations
    """
    values: List = field(default_factory=lambda: [])
    priority: int = 0



# SPDX-FileCopyrightText: Coypright © 2024 Shooting Soul Ventures, LLC <jg@shootingsoul.com>
# SPDX-License-Identifier: MIT

from enum import Enum
from typing import Any, Union, List, Tuple
import itertools
from configsweep.sweep import Sweep
from dataclasses import is_dataclass
from operator import attrgetter


class SweepDag:
    """
    A dag of sweep items from the given config
    Each node in the dag has a list of combinations ot process
    Thus, the root node has all the combionations to process for the entire config

    First the DAG nodes are built
    Second, each child node is sorted according to the priority
    Third, the combinations are built in the priorty order

    NOTE: this is constantly modifying the config passed in 
    and keeps references to objects in the config passed in
    """

    def __init__(self, config: Any):
        self.config = config
        # need a single value for combo algorithm
        self.root_node = SweepDag.Node(Sweep([None]))
        self.build_nodes(self.root_node, 0, "", None, None, None, None, config)
        self.sort_priority(self.root_node)
        self.build_combos(self.root_node)

    class ComboElement:
        def __init__(self, node, value_index: int, next=None):
            self.node: SweepDag.Node = node
            self.value_index: int = value_index
            self.next: SweepDag.ComboElement = next

        def value(self):
            return self.node.values[self.value_index]

        def value_description(self, with_name: bool = False):
            # build description for sweep item replaced
            value = self.value()
            value_desc = _short_value_description(value)
            if not value_desc:
                value_desc = f"<complex_value>[{self.value_index}]"
            return f"{self.node.object_name}={value_desc}" if with_name else value_desc

        def last_value_description(self, with_name: bool = False):
            last = self
            while last.next is not None:
                last = last.next
            return last.value_description(with_name)

    class Node:
        def __init__(self,
                     sweep: Sweep,
                     parent=None,
                     parent_value_index: int = None,
                     object_name: str = None,
                     attribute_name: str = None,
                     key_name: str = None,
                     list_index: int = None,
                     object=None):
            self.sweep = sweep  # original sweep item in the config
            self.parent: SweepDag.Node = parent
            self.parent_value_index: int = parent_value_index
            self.object_name: str = object_name
            self.attribute_name = attribute_name  # also indicates obj is a class
            self.key_name = key_name  # also indicates obj is a dict
            self.list_index = list_index  # also indicates obj is a list
            self.object = object
            self.priority = sweep.priority
            # sweep node values to sweep through must be a non-empty list
            if sweep.values is None or not isinstance(sweep.values, list) or not len(sweep.values):
                raise ValueError(
                    f"{object_name} Sweep values must contain a non-empty list")
            self.values = sweep.values
            self.child_nodes: List[SweepDag.Node] = []
            self.combos: List[List[SweepDag.ComboElement]] = []

    def build_nodes(self, parent, parent_value_index: int, object_name: str, attribute_name: str, key_name: str, list_index: int, object, value):
        """
        Pull out the DAG of sweep nodes from config file
        Traverse through all dicts, lists and dataclasses in the config to extract the sweep node DAG
        The DAG only contains sweep nodes with references back to the original config
        """
        if isinstance(value, Sweep):
            # check if came from a tuple/set.  Can't swap out values in that
            # set/tuple can't contain a sweep object (not hashable), so this should never come up in theory
            if isinstance(object, (set, tuple, frozenset)):
                raise TypeError(
                    f"Can't sweep items in a tuple or set.  See if item {object_name} can be converted to a list.")

            node = SweepDag.Node(value, parent, parent_value_index,
                                 object_name, attribute_name, key_name, list_index, object)
            parent.child_nodes.append(node)
            # support sweep within sweep (could be nested right away or later on)
            # cut out current sweep node, set it as the parent and use it's values to then proceed as normal
            parent = node
            value = value.values
            for idx, next_value in enumerate(value):
                if isinstance(next_value, Sweep):
                    raise TypeError((f"{object_name} Sweep value can't be another sweep directly.  "
                                     "Use one Sweep item with a list of merged values or make an array to sweep an element in the array"))
                self.build_nodes(parent, idx, object_name,
                                 None, None, idx, value, next_value)
        elif isinstance(value, dict):
            for name, next_value in value.items():
                next_object_name = f"{object_name}.{name}" if len(
                    object_name) else name
                self.build_nodes(parent, parent_value_index,
                                 next_object_name, None, name, None, value, next_value)
        elif is_dataclass(value):
            for name, next_value in vars(value).items():
                # skip dunders
                if not name.startswith('__'):
                    next_object_name = f"{object_name}.{name}" if len(
                        object_name) else name
                    self.build_nodes(
                        parent, parent_value_index, next_object_name, name, None, None, value, next_value)
        elif isinstance(value, (list, set, tuple, frozenset)):
            # it's ok to go through tuple/set, but not to sweep items inside a tuple or set
            for idx, next_value in enumerate(value):
                self.build_nodes(parent, parent_value_index,
                                 f"{object_name}[{idx}]", None, None, idx, value, next_value)

    def sort_priority(self, node):
        """
        Sort the children of each node by priority 
        This ensures sweep items are processed in the correct order to keep items together across combinations
        """
        if len(node.child_nodes):
            for child in node.child_nodes:
                self.sort_priority(child)
            node.child_nodes.sort(key=attrgetter('priority'), reverse=True)

    def build_combos(self, node):
        """
        Builds the combinations for each node in the dag with children already sorted correctly

        Handle nested sweeps
        If a sweep index has child sweeps then take the product of those children
        Also keep a list for the combo element to know all the nested replacements to make down stream

        The base case is there are no kids, so just use the sweep values as the combos
        """
        if not len(node.child_nodes):
            # base case optimized for when no kids
            # value is the combo
            node.combos = [[SweepDag.ComboElement(
                node, i)] for i in range(len(node.values))]
            return
        else:
            for child in node.child_nodes:
                self.build_combos(child)

            for value_index in range(len(node.values)):
                index_had_a_child = False
                # when child sweeps share the same parent index
                # then take the product of those children
                a = []
                for child in node.child_nodes:
                    if child.parent_value_index == value_index:
                        sub_list = []
                        for child_combo in child.combos:
                            # need to replace for each child combo current element and the child element
                            sub_list.append([SweepDag.ComboElement(
                                node, value_index, ce) for ce in child_combo])
                        a.append(sub_list)
                        index_had_a_child = True
                if index_had_a_child:
                    p = list(itertools.product(*a))
                    # product gives a tuple of items from each list
                    # however, child elements will be a list of combo elements
                    # flatten it to a list of combo elements
                    for combo_tuple in p:
                        flat = []
                        for i in combo_tuple:
                            # know it will be a plain list from the product function
                            if isinstance(i, list):
                                for j in i:
                                    flat.append(j)
                            else:
                                flat.append(i)
                        node.combos.append(flat)
                else:
                    # base case - append the value as the combo
                    node.combos.append(
                        [SweepDag.ComboElement(node, value_index)])

    def apply_combo_to_config(self, combo_index: int) -> List[str]:
        """
        Apply all the values in the combo to the config
        So we go through the config and swap out all the right values in all the right places

        Each element in the combo has a reference to the config to replace 
        and a list sub elements to replace as well
        """
        combo = self.root_node.combos[combo_index]
        description_list = []
        for element in combo:
            element_description_list = []
            # make all replacements down the chain for nested sweeps
            while element is not None:
                node = element.node
                value = node.values[element.value_index]
                object = node.object
                if object is not None:
                    if node.attribute_name is not None:
                        setattr(object, node.attribute_name, value)
                    elif node.key_name is not None:
                        object[node.key_name] = value
                    else:
                        object[node.list_index] = value
                    element_description_list.append(element.value_description(with_name=True))
                    # if element.next is None:
                    #     # use last element in chain for the description
                    #     element_description_list.append(
                    #         element.value_description(with_name=True))
                    # else:
                    #     element_description_list.append(node.object_name)
                element = element.next
            description_list.append(' & '.join(element_description_list))
        return description_list

    def apply_sweep_to_config(self):
        """
        Put the sweep objects back into the config
        i.e. put it back to how it was originally
        """
        self._apply_sweep(self.root_node)

    def _apply_sweep(self, node):
        if node is None:
            return

        object = node.object
        if object is not None:
            value = node.sweep
            if node.attribute_name is not None:
                setattr(object, node.attribute_name, value)
            elif node.key_name is not None:
                object[node.key_name] = value
            else:
                object[node.list_index] = value

        for child in node.child_nodes:
            self._apply_sweep(child)


def _short_value_description(value: Any) -> Union[str, None]:
    # guess at a short description for the value
    # return None for can't decide/too long
    # put a limit on the size
    LIMIT = 100

    description = None
    if isinstance(value, (str, bool, float, int, Enum)):
        description = str(value)
        if len(description) > LIMIT:
            description = description[:LIMIT - 3] + "..."
    elif isinstance(value, (list, dict, set, tuple)):
        # don't bother for collections
        pass
    else:
        # see if we can get a description of the class that is short
        description = str(value)
        if len(description) > LIMIT:
            description = None  # nevermind, we tried

    return description



# SPDX-FileCopyrightText: Coypright © 2024 Shooting Soul Ventures, LLC <jg@shootingsoul.com>
# SPDX-License-Identifier: MIT

import pytest
from dataclasses import dataclass, field
from configsweep import Sweep, Sweeper
from enum import Enum
from typing import Any
from copy import deepcopy


class MyDayPart(Enum):
    DAWN = 1
    MORNING = 2
    NOON = 3
    AFTERNOON = 4
    DUSK = 5
    NIGHT = 6


@dataclass
class MyMetric:
    min: int = 0
    max: int = 0


@dataclass
class MySillyMetric:
    target: int = 0
    metric: MyMetric = field(default_factory=lambda: MyMetric())


@dataclass
class MyConfig:
    name: str = ""
    day_part: MyDayPart = MyDayPart.AFTERNOON
    top_three: list[str] = field(default_factory=lambda: [
                                 "Miami", "Chicago", "NY"])
    data: dict = field(default_factory=lambda: {})
    metric: Any = None
    metric_grid: Any = None


def test_no_sweep():
    config = MyConfig("racing")
    sweeper = Sweeper(config)
    assert len(sweeper) == 1
    iterator = iter(sweeper)
    combo = next(sweeper)
    assert combo.config.name == "racing"
    with pytest.raises(StopIteration) as e_info:
        next(iterator)


def test_enum():
    config = MyConfig("racing")
    config.day_part = Sweep([MyDayPart.DAWN, None, MyDayPart.DUSK])
    sweeper = Sweeper(config)
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.config.day_part == MyDayPart.DAWN
    combo = next(iterator)
    assert combo.config.day_part is None
    combo = next(iterator)
    assert combo.config.day_part == MyDayPart.DUSK
    with pytest.raises(StopIteration) as e_info:
        next(iterator)


def test_list():
    config = MyConfig("racing", top_three=Sweep(
        [["LA", "NY", "Philly"], ["Maui", "Kauai", "Oahu"], ["1", "2", "3", "4"]]))
    sweeper = Sweeper(config)
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.config.top_three == ["LA", "NY", "Philly"]
    combo = next(iterator)
    assert combo.config.top_three == ["Maui", "Kauai", "Oahu"]
    combo = next(iterator)
    assert combo.config.top_three == ["1", "2", "3", "4"]
    with pytest.raises(StopIteration) as e_info:
        next(iterator)


def test_list_index():
    config = MyConfig("racing", top_three=["LA", Sweep(
        ["Detroit", "Chicago", None, "NY"]), "Philly"])
    sweeper = Sweeper(config)
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.config.top_three == ["LA", "Detroit", "Philly"]
    combo = next(iterator)
    assert combo.config.top_three == ["LA", "Chicago", "Philly"]
    combo = next(iterator)
    assert combo.config.top_three == ["LA", None, "Philly"]
    combo = next(iterator)
    assert combo.config.top_three == ["LA", "NY", "Philly"]
    with pytest.raises(StopIteration) as e_info:
        next(iterator)


def test_dict():
    config = MyConfig("racing", data=Sweep(
        [{"hey": 1, "now": 3}, {"hey": 4, "now": 5}]))
    sweeper = Sweeper(config)
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.config.data == {"hey": 1, "now": 3}
    combo = next(iterator)
    assert combo.config.data == {"hey": 4, "now": 5}
    with pytest.raises(StopIteration) as e_info:
        next(iterator)


def test_dict_attribute():
    config = MyConfig("racing", data={"hey": 1, "now": Sweep([3, 6, 7, 8])})
    sweeper = Sweeper(config)
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.config.data == {"hey": 1, "now": 3}
    combo = next(iterator)
    assert combo.config.data == {"hey": 1, "now": 6}
    combo = next(iterator)
    assert combo.config.data == {"hey": 1, "now": 7}
    combo = next(iterator)
    assert combo.config.data == {"hey": 1, "now": 8}
    with pytest.raises(StopIteration) as e_info:
        next(iterator)


def test_multiple():
    config = MyConfig("racing", top_three=Sweep(
        [["LA", "NY", "Philly"], ["Maui", "Kauai", "Oahu"], ["1", "2", "3", "4"]]))
    config.day_part = Sweep([MyDayPart.DAWN, MyDayPart.DUSK])
    sweeper = Sweeper(config)
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.config.day_part == MyDayPart.DAWN
    assert combo.config.top_three == ["LA", "NY", "Philly"]
    combo = next(iterator)
    assert combo.config.day_part == MyDayPart.DAWN
    assert combo.config.top_three == ["Maui", "Kauai", "Oahu"]
    combo = next(iterator)
    assert combo.config.day_part == MyDayPart.DAWN
    assert combo.config.top_three == ["1", "2", "3", "4"]
    combo = next(iterator)
    assert combo.config.day_part == MyDayPart.DUSK
    assert combo.config.top_three == ["LA", "NY", "Philly"]
    combo = next(iterator)
    assert combo.config.day_part == MyDayPart.DUSK
    assert combo.config.top_three == ["Maui", "Kauai", "Oahu"]
    combo = next(iterator)
    assert combo.config.day_part == MyDayPart.DUSK
    assert combo.config.top_three == ["1", "2", "3", "4"]
    with pytest.raises(StopIteration) as e_info:
        next(iterator)


def test_class():
    config = MyConfig("racing", metric=Sweep(
        [MyMetric(5, 10), MyMetric(100, 200)]))
    sweeper = Sweeper(config)
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.config.metric == MyMetric(5, 10)
    combo = next(iterator)
    assert combo.config.metric == MyMetric(100, 200)
    with pytest.raises(StopIteration) as e_info:
        next(iterator)


def test_class_attribute():
    config = MyConfig("racing", metric=MyMetric(max=5))
    config.metric.min = Sweep([0, 1, 2, 3])
    sweeper = Sweeper(config)
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.config.metric.min == 0
    combo = next(iterator)
    assert combo.config.metric.min == 1
    combo = next(iterator)
    assert combo.config.metric.min == 2
    combo = next(iterator)
    assert combo.config.metric.min == 3
    with pytest.raises(StopIteration) as e_info:
        next(iterator)


def test_value_not_list():
    config = MyConfig("racing", metric=Sweep(123))
    with pytest.raises(ValueError) as e_info:
        Sweeper(config)


def test_one():
    config = MyConfig("racing", metric=Sweep([MyMetric(5, 10)]))
    sweeper = Sweeper(config)
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.config.metric == MyMetric(5, 10)
    with pytest.raises(StopIteration) as e_info:
        next(iterator)


def test_none():
    config = MyConfig("racing", metric=Sweep(None))
    with pytest.raises(ValueError) as e_info:
        Sweeper(config)


def test_metric_grid():
    config = MyConfig("racing", metric_grid=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    config.metric_grid[1][1] = Sweep([1000, 2000])
    sweeper = Sweeper(config)
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.config.metric_grid[2][2] == 9
    assert combo.config.metric_grid[1][1] == 1000
    combo = next(iterator)
    assert combo.config.metric_grid[2][2] == 9
    assert combo.config.metric_grid[1][1] == 2000
    with pytest.raises(StopIteration) as e_info:
        next(iterator)


def test_empty_list():
    config = MyConfig("racing", metric=Sweep([]))
    with pytest.raises(ValueError) as e_info:
        Sweeper(config)


def test_sweep_through_tuple():

    # can't swap in a set or tuple, but can further down the chain
    # make sure we traverse set / tuple
    # set/tuple can't be built with sweep, so should be good
    config = {"a": tuple([{"x": 123}, {"y": Sweep([10, 20, 30])}, "c"])}
    sweeper = Sweeper(config)
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.config['a'][1]['y'] == 10
    combo = next(iterator)
    assert combo.config['a'][1]['y'] == 20
    combo = next(iterator)
    assert combo.config['a'][1]['y'] == 30
    with pytest.raises(StopIteration) as e_info:
        next(iterator)


def test_sweep_sweep():
    # test direct nested - not allowed for ambiguity
    # could want to sweep an element in the array or just have a flat array

    # for example
    # config = {"interval": "15", "strategy": {"name": "s1", "max": Sweep([Sweep([10000,90000]),10])}}
    # could want this:
    # config = {"interval": "15", "strategy": {"name": "s1", "max": Sweep([[Sweep([10000,90000]),10]])}}
    # - or could want this-
    # config = {"interval": "15", "strategy": {"name": "s1", "max": Sweep([10000,90000,10])}}

    config = {"interval": "15", "strategy": {
        "name": "s1", "max": Sweep([Sweep([10000, 90000]), 10])}}
    with pytest.raises(TypeError) as e_info:
        sweeper = Sweeper(config)

# copy option not available . . .
# def test_copy():
#     config = {"x": 123, "y": Sweep([10,20,30]), "c": 456 }
#     original_config = deepcopy(config)

#     # first time, let sweeper make a copy and verify config still matches original
#     sweeper = Sweeper(config)
#     iterator = iter(sweeper)
#     combo = next(iterator)
#     assert combo.config['y'] == 10
#     combo = next(iterator)
#     assert combo.config['y'] == 20
#     combo = next(iterator)
#     assert combo.config['y'] == 30
#     with pytest.raises(StopIteration) as e_info:
#         next(iterator)
#     assert config == original_config

#     # second time, no copy, make sure sweep still works
#     # and config matches original after done sweeping
#     # config won't match original while sweeping
#     sweeper = Sweeper(config, copy = False)
#     sweeper._copy
#     iterator = iter(sweeper)
#     combo = next(iterator)
#     assert combo.config['y'] == 10
#     combo = next(iterator)
#     assert combo.config['y'] == 20
#     combo = next(iterator)
#     assert combo.config['y'] == 30
#     with pytest.raises(StopIteration) as e_info:
#         next(iterator)
#     assert config == original_config


def test_multiple_configs():
    config = {"x": 123, "y": Sweep([10, 20, 30]), "c": 456}
    original_config = deepcopy(config)

    config_racing = MyConfig("racing")
    config_racing.day_part = Sweep([MyDayPart.DAWN, None, MyDayPart.DUSK])
    original_config_racing = deepcopy(config_racing)

    sweeper = Sweeper([config, config_racing])
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.config['y'] == 10
    combo = next(iterator)
    assert combo.config['y'] == 20
    combo = next(iterator)
    assert combo.config['y'] == 30

    combo = next(iterator)
    assert combo.config.day_part == MyDayPart.DAWN
    combo = next(iterator)
    assert combo.config.day_part is None
    combo = next(iterator)
    assert combo.config.day_part == MyDayPart.DUSK
    with pytest.raises(StopIteration) as e_info:
        next(iterator)

    assert config == original_config
    assert config_racing == original_config_racing

def test_nested_priority():
    config = { "strategy": Sweep([
               {"name": "strategy_one", "max": 10000},
               {"name": "strategy_two", "min": Sweep([10,20,30]), "max": Sweep([10000,90000])}
               ]),
            "datasources": Sweep(["en", "es", "de", "fr"], priority=1)
        }

    sweeper = Sweeper(config)
    iterator = iter(sweeper)
    combo = next(iterator)
    assert combo.description == "datasources=en\nstrategy=<complex_value>[0]"
    combo = next(iterator)
    assert combo.description == "datasources=en\nstrategy=<complex_value>[1] & strategy.min=10\nstrategy=<complex_value>[1] & strategy.max=10000"



# SPDX-FileCopyrightText: Coypright © 2024 Shooting Soul Ventures, LLC <jg@shootingsoul.com>
# SPDX-License-Identifier: MIT

from typing import ClassVar, Type


class AffiliateClass:
    """
    Mixin for use with a config dataclass.

    This associates the class being configured with the config for easy creation
    """

    _affiliate_cls: ClassVar[Type] = None

    def __init__():
        pass

    def create_affiliate(self, *args, **kwargs):
        """
        Create an instance of the affiliate class for the config.

        This config class is passed as the first argument to the affiliate contstructor
        along with any additional args, kwargs
        """
        if self._affiliate_cls is None:
            raise ValueError(
                f"Affiliate class not set.  Missing the assignment {type(self).__name__}._affiliate_cls = <type>")
        return self._affiliate_cls(self, *args, **kwargs)



# SPDX-FileCopyrightText: Coypright © 2024 Shooting Soul Ventures, LLC <jg@shootingsoul.com>
# SPDX-License-Identifier: MIT

import pytest
from dataclasses import dataclass
from configsweep.affiliate import AffiliateClass
from classifiedjson import dumps, loads


@dataclass
class MyConfig(AffiliateClass):
    max: int = None
    min: int = -100


class MySubject:
    def __init__(self, config: MyConfig):
        self.config = config
        self.limit = 5 * config.max


MyConfig._affiliate_cls = MySubject


@dataclass
class MyConfigMissing(AffiliateClass):
    avg: int = 99
    mean: int = 78


def test_basic():
    c = MyConfig(max=1000)
    s = c.create_affiliate()
    assert s.config.min == -100
    assert s.config.max == 1000


def test_serialized():
    config = MyConfig(max=2000)
    s = dumps(config)
    config2 = loads(s)
    assert config == config2
    subject2 = config2.create_affiliate()
    assert subject2.limit == 10000


def test_missing():
    config = MyConfigMissing()
    # forgot to set _affiliate_cls.
    # catch this on attemp to create it
    with pytest.raises(ValueError) as e_info:
        config.create_affiliate()



# SPDX-FileCopyrightText: Coypright © 2024 Shooting Soul Ventures, LLC <jg@shootingsoul.com>
# SPDX-License-Identifier: MIT

from typing import Any


class SweepCombination:
    """
    A config with values set for a combination of sweep items defined in the original config

    description - what values where set for this combination
    config - the config with all sweep items replaced with values
    index - a unique index across all the sweep combinations
    """

    def __init__(self, description: str, config: Any, index: int):
        # description of the sweep values used in this combination
        self.description = description
        self.config = config
        self.index = index

    def __repr__(self) -> str:
        return f"SweepCombination(index={self.index} description={self.description}, config={self.config})"



# SPDX-FileCopyrightText: Coypright © 2024 Shooting Soul Ventures, LLC <jg@shootingsoul.com>
# SPDX-License-Identifier: MIT

from copy import deepcopy
from configsweep.sweep_dag import SweepDag
from configsweep.sweep_combination import SweepCombination
from typing import Any, List, Union


class Sweeper:
    """
    Iterator sweeps over a config with Sweep objects
    Each iteration replaces each sweep object with a combinations of values
    Nested sweep objects are supported
    
    NOTE: a copy of the config is made and copies of the config for each combination are made
    """

    def __init__(self, config: Union[Any, List]):
        if config is None:
            config_list = []
        else:
            config_list = config if isinstance(config, list) else [config]

        # create a copy of the config to hack up and sweep through the Sweep class value list combos
        # if no copy is made:
        #   Config is shared for all combos and put back in place after all sweeps
        #   This means each sweepcombination.config is lost on the next iteration
        #   Exiting the sweeper before completion all combinations will leave config in a bad state
        # So we won't expose this option for now
        self._copy = True
        self._config_templates = [
            deepcopy(s) for s in config_list] if self._copy else config_list
        self._dags = [SweepDag(s) for s in self._config_templates]
        self._len = 0
        for d in self._dags:
            self._len += len(d.root_node.combos)
        self._current_dag = None

    def __len__(self) -> int:
        return self._len

    def __iter__(self):
        self._pos = 0
        self._dag_index = 0
        self._combo_index = 0
        if self._len:
            self._current_dag = self._dags[self._dag_index]
        else:
            self._current_dag = None
        return self

    def __next__(self) -> SweepCombination:
        # beating a dead horse
        if self._current_dag is None:
            raise StopIteration

        # see if we are done with combos for the current dag
        if self._combo_index == len(self._current_dag.root_node.combos):
            if not self._copy:
                # set config back the way it was to start with the sweep objects in place
                # only needed if not copying
                self._current_dag.apply_sweep_to_config()
            # on to the next config/dag if available
            self._dag_index += 1
            self._combo_index = 0  # reset pos to iterate through next dag
            if self._dag_index == len(self._dags):
                self._current_dag = None  # fin
                raise StopIteration
            else:
                self._current_dag = self._dags[self._dag_index]

        # apply the combo to substitute values in the config
        # get a description of the substitutions made
        description_list = self._current_dag.apply_combo_to_config(
            self._combo_index)

        # return a copy of the config with all sweep values replaced
        # also include description of the sweep combination used
        if self._copy:
            config = deepcopy(self._config_templates[self._dag_index])
        else:
            config = self._config_templates[self._dag_index]

        combo = SweepCombination(
            "\n".join(description_list), config, self._pos)
        # prep for next combo
        self._combo_index += 1
        self._pos += 1
        return combo



# SPDX-FileCopyrightText: Coypright © 2024 Shooting Soul Ventures, LLC <jg@shootingsoul.com>
# SPDX-License-Identifier: MIT

from configsweep.affiliate import AffiliateClass
from configsweep.sweep import Sweep
from configsweep.sweep_combination import SweepCombination
from configsweep.sweeper import Sweeper

__version__ = "1.0.0"
__all__ = (__version__,
           AffiliateClass,
           Sweep,
           Sweeper,
           SweepCombination)
